use std::cell::RefCell;
use std::env;
use std::process::{Command, Stdio};
use std::rc::Rc;

use eyre::{bail, Result, WrapErr};

use super::pattern::r#match;
use super::r#macro::{Macro, Set as MacroSet, Source as MacroSource};
use super::token::TokenString;

pub fn expand_call(
    name: &str,
    args: &[TokenString],
    macros: &MacroSet,
    to_eval: Option<Rc<RefCell<Vec<String>>>>,
) -> Result<String> {
    match name {
        "subst" => {
            assert_eq!(args.len(), 3);
            text::subst(macros, &args[0], &args[1], &args[2])
        }
        "patsubst" => {
            assert_eq!(args.len(), 3);
            text::patsubst(macros, &args[0], &args[1], &args[2])
        }
        "strip" => {
            assert_eq!(args.len(), 1);
            text::strip(macros, &args[0])
        }
        "findstring" => {
            assert_eq!(args.len(), 2);
            text::findstring(macros, &args[0], &args[1])
        }
        "filter" => {
            assert_eq!(args.len(), 2);
            text::filter(macros, &args[0], &args[1])
        }
        "filter-out" => {
            assert_eq!(args.len(), 2);
            text::filter_out(macros, &args[0], &args[1])
        }
        "sort" => {
            assert_eq!(args.len(), 1);
            text::sort(macros, &args[0])
        }
        "word" => {
            assert_eq!(args.len(), 2);
            text::word(macros, &args[0], &args[1])
        }
        "words" => {
            assert_eq!(args.len(), 1);
            text::words(macros, &args[0])
        }
        "lastword" => {
            assert_eq!(args.len(), 1);
            text::lastword(macros, &args[0])
        }

        "dir" => {
            assert_eq!(args.len(), 1);
            file_name::dir(macros, &args[0])
        }
        "notdir" => {
            assert_eq!(args.len(), 1);
            file_name::notdir(macros, &args[0])
        }
        "basename" => {
            assert_eq!(args.len(), 1);
            file_name::basename(macros, &args[0])
        }
        "addsuffix" => {
            assert_eq!(args.len(), 2);
            file_name::addsuffix(macros, &args[0], &args[1])
        }
        "addprefix" => {
            assert_eq!(args.len(), 2);
            file_name::addprefix(macros, &args[0], &args[1])
        }
        "wildcard" => {
            assert_eq!(args.len(), 1);
            file_name::wildcard(macros, &args[0])
        }
        "realpath" => {
            assert_eq!(args.len(), 1);
            file_name::realpath(macros, &args[0])
        }

        "if" => {
            assert!(args.len() == 2 || args.len() == 3);
            conditional::r#if(macros, &args[0], &args[1], args.get(2))
        }
        "or" => {
            assert!(!args.is_empty());
            conditional::or(macros, args.iter())
        }
        "and" => {
            assert!(!args.is_empty());
            conditional::and(macros, args.iter())
        }

        "foreach" => {
            assert_eq!(args.len(), 3);
            foreach(macros, &args[0], &args[1], &args[2])
        }

        "call" => {
            assert!(!args.is_empty());
            call(macros, args.iter())
        }

        "eval" => {
            assert_eq!(args.len(), 1);
            let should_eval = eval(macros, &args[0])?;
            if let Some(to_eval) = to_eval {
                to_eval.borrow_mut().push(should_eval);
            } else {
                bail!("tried to eval something but no eval back-channel was available");
            }
            Ok(String::new())
        }

        "origin" => {
            assert_eq!(args.len(), 1);
            origin(macros, &args[0])
        }

        "error" => {
            assert_eq!(args.len(), 1);
            meta::error(macros, &args[0])
        }

        "shell" => {
            assert_eq!(args.len(), 1);
            shell(macros, &args[0])
        }

        // fallback
        _ => bail!("function not implemented: {}", name),
    }
}

// Text Functions
mod text {
    use super::*;

    pub fn subst(
        macros: &MacroSet,
        from: &TokenString,
        to: &TokenString,
        text: &TokenString,
    ) -> Result<String> {
        let from = macros.expand(from)?;
        let to = macros.expand(to)?;
        let text = macros.expand(text)?;
        Ok(text.replace(&from, &to))
    }

    pub fn patsubst(
        macros: &MacroSet,
        from: &TokenString,
        to: &TokenString,
        text: &TokenString,
    ) -> Result<String> {
        let from = macros.expand(from)?;
        let to = macros.expand(to)?;
        let text = macros.expand(text)?;
        let words = text
            .split_whitespace()
            .map(|word| {
                let pattern_match = r#match(&from, word)?.and_then(|x| x.get(1));
                Ok(if let Some(pm) = pattern_match {
                    to.replace('%', pm.as_str())
                } else {
                    word.to_owned()
                })
            })
            .collect::<Result<Vec<_>>>()?;
        Ok(words.join(" "))
    }

    pub fn strip(macros: &MacroSet, text: &TokenString) -> Result<String> {
        let text = macros.expand(text)?;
        // TODO don't allocate this vec
        let words = text.split_whitespace().collect::<Vec<_>>();
        Ok(words.join(" "))
    }

    pub fn findstring(
        macros: &MacroSet,
        needle: &TokenString,
        haystack: &TokenString,
    ) -> Result<String> {
        let needle = macros.expand(needle)?;
        let haystack = macros.expand(haystack)?;
        if haystack.contains(&needle) {
            Ok(needle)
        } else {
            Ok(String::new())
        }
    }

    pub fn filter(macros: &MacroSet, patterns: &TokenString, text: &TokenString) -> Result<String> {
        let patterns = macros.expand(patterns)?;
        let patterns = patterns.split_whitespace().collect::<Vec<_>>();
        let text = macros.expand(text)?;
        let text = text.split_whitespace();
        let mut result_pieces = vec![];
        for word in text {
            if patterns
                .iter()
                .any(|pattern| r#match(pattern, word).map_or(false, |x| x.is_some()))
            {
                result_pieces.push(word);
            }
        }
        Ok(result_pieces.join(" "))
    }

    pub fn filter_out(
        macros: &MacroSet,
        patterns: &TokenString,
        text: &TokenString,
    ) -> Result<String> {
        let patterns = macros.expand(patterns)?;
        let patterns = patterns.split_whitespace().collect::<Vec<_>>();
        let text = macros.expand(text)?;
        let text = text.split_whitespace();
        let mut result_pieces = vec![];
        for word in text {
            if patterns
                .iter()
                .all(|pattern| r#match(pattern, word).map_or(false, |x| x.is_none()))
            {
                result_pieces.push(word);
            }
        }
        Ok(result_pieces.join(" "))
    }

    pub fn sort(macros: &MacroSet, words: &TokenString) -> Result<String> {
        let words = macros.expand(words)?;
        let mut words = words.split_whitespace().collect::<Vec<_>>();
        words.sort_unstable();
        words.dedup();
        Ok(words.join(" "))
    }

    pub fn word(macros: &MacroSet, n: &TokenString, text: &TokenString) -> Result<String> {
        let n = macros.expand(n)?;
        let n: usize = n.parse().wrap_err("while calling `word`")?;
        let text = macros.expand(text)?;
        Ok(text
            .split_whitespace()
            .nth(n.saturating_add(1))
            .unwrap_or("")
            .to_owned())
    }

    pub fn words(macros: &MacroSet, words: &TokenString) -> Result<String> {
        let words = macros.expand(words)?;
        Ok(words.split_whitespace().count().to_string())
    }

    pub fn lastword(macros: &MacroSet, words: &TokenString) -> Result<String> {
        let words = macros.expand(words)?;
        Ok(words.split_whitespace().last().unwrap_or("").to_owned())
    }
}

// File Name Functions
mod file_name {
    use std::env;
    use std::ffi::OsStr;
    use std::fs;
    use std::path::Path;

    use eyre::WrapErr;

    use super::*;

    pub fn dir(macros: &MacroSet, words: &TokenString) -> Result<String> {
        let words = macros.expand(words)?;
        let words = words
            .split_whitespace()
            .map(|word| {
                Path::new(word)
                    .parent()
                    .and_then(Path::to_str)
                    .filter(|x| !x.is_empty())
                    .unwrap_or("./")
            })
            .collect::<Vec<_>>();
        Ok(words.join(" "))
    }

    pub fn notdir(macros: &MacroSet, words: &TokenString) -> Result<String> {
        let words = macros.expand(words)?;
        let words = words
            .split_whitespace()
            .map(|word| {
                Path::new(word)
                    .file_name()
                    .and_then(OsStr::to_str)
                    .unwrap_or("")
            })
            .collect::<Vec<_>>();
        Ok(words.join(" "))
    }

    pub fn basename(macros: &MacroSet, words: &TokenString) -> Result<String> {
        let words = macros.expand(words)?;
        let words = words
            .split_whitespace()
            .map(|word| {
                Path::new(word)
                    .with_extension("")
                    .to_str()
                    .map_or_else(String::new, ToString::to_string)
            })
            .collect::<Vec<_>>();
        Ok(words.join(" "))
    }

    pub fn addsuffix(
        macros: &MacroSet,
        suffix: &TokenString,
        targets: &TokenString,
    ) -> Result<String> {
        let suffix = macros.expand(suffix)?;
        let targets = macros.expand(targets)?;
        let results = targets
            .split_whitespace()
            .map(|t| format!("{}{}", t, suffix))
            .collect::<Vec<_>>();
        Ok(results.join(" "))
    }

    pub fn addprefix(
        macros: &MacroSet,
        prefix: &TokenString,
        targets: &TokenString,
    ) -> Result<String> {
        let prefix = macros.expand(prefix)?;
        let targets = macros.expand(targets)?;
        let results = targets
            .split_whitespace()
            .map(|t| format!("{}{}", prefix, t))
            .collect::<Vec<_>>();
        Ok(results.join(" "))
    }

    pub fn wildcard(macros: &MacroSet, pattern: &TokenString) -> Result<String> {
        let pattern = macros.expand(pattern)?;
        let home_dir = env::var("HOME")
            .ok()
            .or_else(|| dirs::home_dir().and_then(|p| p.to_str().map(String::from)));
        let pattern = if let Some(home_dir) = home_dir {
            pattern.replace('~', &home_dir)
        } else {
            pattern
        };
        let results = glob::glob(&pattern)
            .context("invalid glob pattern!")?
            .filter_map(|path| {
                path.ok()
                    .map(|x| x.to_str().map(ToString::to_string).unwrap_or_default())
            })
            .collect::<Vec<_>>();
        Ok(results.join(" "))
    }

    pub fn realpath(macros: &MacroSet, targets: &TokenString) -> Result<String> {
        let targets = macros.expand(targets)?;
        let results = targets
            .split_whitespace()
            .map(|x| {
                fs::canonicalize(x)
                    .map(|p| p.to_string_lossy().into_owned())
                    .unwrap_or_else(|_| x.to_owned())
            })
            .collect::<Vec<_>>();
        Ok(results.join(" "))
    }
}

// Functions for Conditionals
mod conditional {
    use super::*;

    pub fn r#if(
        macros: &MacroSet,
        condition: &TokenString,
        if_true: &TokenString,
        if_false: Option<&TokenString>,
    ) -> Result<String> {
        let mut condition = condition.clone();
        condition.trim_start();
        condition.trim_end();
        let condition = macros.expand(&condition)?;
        if condition.is_empty() {
            if let Some(if_false) = if_false {
                macros.expand(if_false)
            } else {
                Ok(String::new())
            }
        } else {
            macros.expand(if_true)
        }
    }

    pub fn or<'a>(
        macros: &MacroSet,
        args: impl Iterator<Item = &'a TokenString>,
    ) -> Result<String> {
        for arg in args {
            let arg = macros.expand(arg)?;
            if !arg.is_empty() {
                return Ok(arg);
            }
        }
        Ok(String::new())
    }

    pub fn and<'a>(
        macros: &MacroSet,
        args: impl Iterator<Item = &'a TokenString>,
    ) -> Result<String> {
        let mut last = String::new();
        for arg in args {
            last = macros.expand(arg)?;
            if last.is_empty() {
                return Ok(String::new());
            }
        }
        Ok(last)
    }
}

pub fn foreach(
    macros: &MacroSet,
    var: &TokenString,
    list: &TokenString,
    text: &TokenString,
) -> Result<String> {
    let var = macros.expand(var)?;
    let list = macros.expand(list)?;
    let words = list.split_whitespace();

    let mut macros = macros.with_overlay();
    let results = words
        .map(|word| {
            macros.set(
                var.clone(),
                Macro {
                    source: MacroSource::File,
                    text: TokenString::text(word),
                    #[cfg(feature = "full")]
                    eagerly_expanded: false,
                },
            );
            macros.expand(text)
        })
        .collect::<Result<Vec<_>, _>>()?;
    Ok(results.join(" "))
}

pub fn call<'a>(macros: &MacroSet, args: impl Iterator<Item = &'a TokenString>) -> Result<String> {
    let args = args
        .map(|arg| macros.expand(arg))
        .collect::<Result<Vec<_>, _>>()?;
    let function = args[0].clone();
    // TODO if function is a builtin, call the builtin instead

    let mut macros = macros.with_overlay();
    for (i, x) in args.into_iter().enumerate() {
        macros.set(
            i.to_string(),
            Macro {
                source: MacroSource::File,
                text: TokenString::text(x),
                #[cfg(feature = "full")]
                eagerly_expanded: false,
            },
        );
    }
    macros.expand(&TokenString::r#macro(function))
}

// TODO consider bringing eval logic in here since we put the Vec in MacroSet IIRC
pub fn eval(macros: &MacroSet, arg: &TokenString) -> Result<String> {
    macros.expand(arg)
}

pub fn origin(macros: &MacroSet, variable: &TokenString) -> Result<String> {
    let variable = macros.expand(variable)?;
    Ok(macros.origin(&variable).to_owned())
}

mod meta {
    use super::*;

    pub fn error(macros: &MacroSet, text: &TokenString) -> Result<String> {
        let text = macros.expand(text)?;
        bail!("{}", text);
    }
}

pub fn shell(macros: &MacroSet, command: &TokenString) -> Result<String> {
    // TODO bring this in from command_line
    let command = macros.expand(command)?;
    let (program, args) = if cfg!(windows) {
        let cmd = env::var("COMSPEC").unwrap_or_else(|_| "cmd.exe".into());
        let args = vec!["/c", &command];
        (cmd, args)
    } else {
        let sh = env::var("SHELL").unwrap_or_else(|_| "/bin/sh".into());
        let args = vec!["-e", "-c", &command];
        (sh, args)
    };
    let result = Command::new(program)
        .args(args)
        .stderr(Stdio::inherit())
        .output()?;
    let _status = result.status;
    // TODO set .SHELLSTATUS
    Ok(String::from_utf8(result.stdout)?
        .replace("\r\n", "\n")
        .trim_end_matches('\n')
        .replace("\n", " "))
}

#[cfg(test)]
mod test {
    use super::*;

    use crate::makefile::r#macro::{Set as MacroSet, Source as MacroSource};

    type R = Result<()>;

    fn call(name: &str, args: &[TokenString], macros: &MacroSet) -> Result<String> {
        super::expand_call(name, args, macros, None)
    }

    macro_rules! call {
        ($func:literal $($arg:literal),+) => {
            call($func, &[$(TokenString::text($arg)),+], &MacroSet::new())?
        };
        ($func:ident $($arg:literal),+) => {
            call(stringify!($func), &[$(TokenString::text($arg)),+], &MacroSet::new())?
        };
    }

    #[test]
    fn strip() -> R {
        let result = call!(strip "  this      is\tweirdly     spaced  text      ");
        assert_eq!(result, "this is weirdly spaced text");
        Ok(())
    }

    #[test]
    fn findstring() -> R {
        assert_eq!(call!(findstring "hi", "thighs"), "hi");
        assert_eq!(call!(findstring "hey", "hello there"), "");
        Ok(())
    }

    #[test]
    fn filter() -> R {
        let result = call!(filter "word", "this contains a word inside it");
        assert_eq!(result, "word");

        let result = call!(filter "%.c %.s", "foo.c bar.c baz.s ugh.h");
        assert_eq!(result, "foo.c bar.c baz.s");
        Ok(())
    }

    #[test]
    fn filter_out() -> R {
        let result = call!("filter-out" "main1.o main2.o", "main1.o foo.o main2.o bar.o");
        assert_eq!(result, "foo.o bar.o");
        Ok(())
    }

    #[test]
    fn sort() -> R {
        let result = call!(sort "foo bar lose foo");
        assert_eq!(result, "bar foo lose");
        Ok(())
    }

    #[test]
    fn notdir() -> R {
        let result = call!(notdir "src/foo.c hacks");
        assert_eq!(result, "foo.c hacks");
        Ok(())
    }

    #[test]
    fn basename() -> R {
        let result = call!(basename "src/foo.c src-1.0/bar hacks");
        assert_eq!(result, "src/foo src-1.0/bar hacks");
        Ok(())
    }

    #[test]
    fn addprefix() -> R {
        let result = call!(addprefix "src/", "foo bar");
        assert_eq!(result, "src/foo src/bar");
        Ok(())
    }

    #[test]
    fn wildcard() -> R {
        use std::env::{set_current_dir, set_var};
        use std::fs::write;
        use std::path::MAIN_SEPARATOR;

        let tempdir = tempfile::tempdir()?;

        write(tempdir.path().join("foo.c"), "")?;
        write(tempdir.path().join("bar.h"), "")?;
        write(tempdir.path().join("baz.txt"), "")?;
        write(tempdir.path().join("acab.c"), "ACAB")?;
        write(tempdir.path().join("based.txt"), "☭")?;

        set_current_dir(tempdir.path())?;
        set_var("HOME", tempdir.path());
        let sort = |x: String| call("sort", &[TokenString::text(&x)], &MacroSet::new());
        assert_eq!(sort(call!(wildcard "*.c"))?, "acab.c foo.c");
        assert_eq!(
            sort(call!(wildcard "~/ba?.*"))?,
            format!(
                "{0}{1}bar.h {0}{1}baz.txt",
                tempdir.path().display(),
                MAIN_SEPARATOR
            )
        );
        Ok(())
    }

    #[test]
    fn test_if() -> R {
        let mut macros = MacroSet::new();
        macros.set(
            "test1".to_owned(),
            Macro {
                source: MacroSource::File,
                text: TokenString::text("something"),
                #[cfg(feature = "full")]
                eagerly_expanded: false,
            },
        );
        macros.set(
            "test2".to_owned(),
            Macro {
                source: MacroSource::File,
                text: TokenString::text(""),
                #[cfg(feature = "full")]
                eagerly_expanded: false,
            },
        );

        assert_eq!(
            call(
                "if",
                &[
                    TokenString::r#macro("test1"),
                    TokenString::text("success"),
                    TokenString::r#macro("failed"),
                ],
                &macros
            )?,
            "success"
        );
        assert_eq!(
            call(
                "if",
                &[
                    TokenString::r#macro("test2"),
                    "$(error failed)".parse()?,
                    TokenString::text("pass"),
                ],
                &macros
            )?,
            "pass"
        );
        Ok(())
    }

    #[test]
    fn or() -> R {
        assert_eq!(
            call(
                "or",
                &[TokenString::text("yep"), TokenString::text("yeah")],
                &MacroSet::new()
            )?,
            "yep"
        );
        assert_eq!(
            call(
                "or",
                &[
                    TokenString::text(""),
                    TokenString::text("yeet"),
                    "$(error fail)".parse()?
                ],
                &MacroSet::new()
            )?,
            "yeet"
        );

        Ok(())
    }

    #[test]
    fn and() -> R {
        assert_eq!(
            call(
                "and",
                &[TokenString::text("yep"), TokenString::text("yeah")],
                &MacroSet::new()
            )?,
            "yeah"
        );
        assert_eq!(
            call(
                "and",
                &[
                    TokenString::text("maybe"),
                    TokenString::text(""),
                    "$(error fail)".parse()?
                ],
                &MacroSet::new()
            )?,
            ""
        );

        Ok(())
    }

    #[test]
    fn foreach() -> R {
        let mut macros = MacroSet::new();
        macros.set(
            "test".to_owned(),
            Macro {
                source: MacroSource::File,
                text: "worked for $(item).".parse()?,
                #[cfg(feature = "full")]
                eagerly_expanded: false,
            },
        );
        assert_eq!(
            call(
                "foreach",
                &[
                    TokenString::text("item"),
                    TokenString::text("a b c d"),
                    TokenString::r#macro("test")
                ],
                &macros,
            )?,
            "worked for a. worked for b. worked for c. worked for d."
        );
        Ok(())
    }

    #[test]
    fn call_test() -> R {
        let mut macros = MacroSet::new();
        macros.set(
            "reverse".to_owned(),
            Macro {
                source: MacroSource::File,
                text: "$(2) $(1)".parse()?,
                #[cfg(feature = "full")]
                eagerly_expanded: false,
            },
        );
        assert_eq!(
            call(
                "call",
                &[
                    TokenString::text("reverse"),
                    TokenString::text("a"),
                    TokenString::text("b")
                ],
                &macros,
            )?,
            "b a"
        );
        Ok(())
    }

    #[test]
    fn shell() -> R {
        let result = call!(shell "echo hi");
        assert_eq!(result, "hi");
        Ok(())
    }
}
